@@ -1,13 +1,14 @@
 """
-HTTPie - cURL for humans.
+HTTPie - a CLI, cURL-like tool for humans.
 
 """
 __author__ = 'Jakub Roztocil'
-__version__ = '0.3.1'
+__version__ = '0.7.1'
 __licence__ = 'BSD'
 
 
-class exit:
+class ExitStatus:
+    """Exit status code constants."""
     OK = 0
     ERROR = 1
     ERROR_TIMEOUT = 2
