@@ -19,25 +19,31 @@ To make it run faster and offline you can::
     HTTPBIN_URL=http://localhost:5000 tox
 
 """
-from functools import partial
 import subprocess
 import os
 import sys
 import json
+#noinspection PyCompatibility
 import argparse
 import tempfile
 import unittest
 import shutil
+import time
+from requests.structures import CaseInsensitiveDict
 
-from requests.compat import urlparse
 try:
+    #noinspection PyCompatibility
     from urllib.request import urlopen
 except ImportError:
+    # noinspection PyUnresolvedReferences
+    #noinspection PyCompatibility
     from urllib2 import urlopen
 try:
     from unittest import skipIf, skip
 except ImportError:
     skip = lambda msg: lambda self: None
+
+    # noinspection PyUnusedLocal
     def skipIf(cond, reason):
         def decorator(test_method):
             if cond:
@@ -46,7 +52,6 @@ except ImportError:
         return decorator
 
 from requests import __version__ as requests_version
-from requests.compat import is_windows, is_py26, bytes, str
 
 
 #################################################################
@@ -57,12 +62,21 @@ from requests.compat import is_windows, is_py26, bytes, str
 TESTS_ROOT = os.path.abspath(os.path.dirname(__file__))
 sys.path.insert(0, os.path.realpath(os.path.join(TESTS_ROOT, '..')))
 
-from httpie import exit
+from httpie import ExitStatus
 from httpie import input
 from httpie.models import Environment
 from httpie.core import main
 from httpie.output import BINARY_SUPPRESSED_NOTICE
 from httpie.input import ParseError
+from httpie.compat import is_windows, is_py26, bytes, str
+from httpie.downloads import (
+    parse_content_range,
+    filename_from_content_disposition,
+    filename_from_url,
+    get_unique_filename,
+    ContentRangeError,
+    Download,
+)
 
 
 CRLF = '\r\n'
@@ -99,7 +113,8 @@ with open(BIN_FILE_PATH, 'rb') as f:
 
 
 def httpbin(path):
-    return HTTPBIN_URL + path
+    url = HTTPBIN_URL + path
+    return url
 
 
 def mk_config_dir():
@@ -132,6 +147,7 @@ class TestEnvironment(Environment):
         if self.delete_config_dir:
             self._shutil.rmtree(self.config_dir)
 
+
 def has_docutils():
     try:
         #noinspection PyUnresolvedReferences
@@ -140,6 +156,7 @@ def has_docutils():
     except ImportError:
         return False
 
+
 def get_readme_errors():
     p = subprocess.Popen([
         'rst2pseudoxml.py',
@@ -154,6 +171,8 @@ def get_readme_errors():
 
 class BytesResponse(bytes):
     stderr = json = exit_status = None
+
+
 class StrResponse(str):
     stderr = json = exit_status = None
 
@@ -177,20 +196,25 @@ def http(*args, **kwargs):
     if not env:
         env = kwargs['env'] = TestEnvironment()
 
+    stdout = env.stdout
+    stderr = env.stderr
     try:
 
         try:
             exit_status = main(args=['--traceback'] + list(args), **kwargs)
+            if '--download' in args:
+                # Let the progress reporter thread finish.
+                time.sleep(.5)
         except Exception:
-            sys.stderr.write(env.stderr.read())
+            sys.stderr.write(stderr.read())
             raise
         except SystemExit:
-            exit_status = exit.ERROR
+            exit_status = ExitStatus.ERROR
 
-        env.stdout.seek(0)
-        env.stderr.seek(0)
+        stdout.seek(0)
+        stderr.seek(0)
 
-        output = env.stdout.read()
+        output = stdout.read()
         try:
             r = StrResponse(output.decode('utf8'))
         except UnicodeDecodeError:
@@ -212,18 +236,20 @@ def http(*args, **kwargs):
                         except ValueError:
                             pass
 
-        r.stderr = env.stderr.read()
+        r.stderr = stderr.read()
         r.exit_status = exit_status
 
         return r
 
     finally:
-        env.stdout.close()
-        env.stderr.close()
+        stdout.close()
+        stderr.close()
 
 
 class BaseTestCase(unittest.TestCase):
 
+    maxDiff = 100000
+
     if is_py26:
         def assertIn(self, member, container, msg=None):
             self.assertTrue(member in container, msg)
@@ -235,11 +261,15 @@ class BaseTestCase(unittest.TestCase):
             self.assertEqual(set(d1.keys()), set(d2.keys()), msg)
             self.assertEqual(sorted(d1.values()), sorted(d2.values()), msg)
 
+    def assertIsNone(self, obj, msg=None):
+        self.assertEqual(obj, None, msg=msg)
+
 
 #################################################################
 # High-level tests using httpbin.
 #################################################################
 
+
 class HTTPieTest(BaseTestCase):
 
     def test_GET(self):
@@ -263,7 +293,7 @@ class HTTPieTest(BaseTestCase):
             'foo=bar'
         )
         self.assertIn(OK, r)
-        self.assertIn('"foo": "bar"', r)
+        self.assertIn(r'\"foo\": \"bar\"', r)
 
     def test_POST_JSON_data(self):
         r = http(
@@ -272,7 +302,7 @@ class HTTPieTest(BaseTestCase):
             'foo=bar'
         )
         self.assertIn(OK, r)
-        self.assertIn('"foo": "bar"', r)
+        self.assertIn(r'\"foo\": \"bar\"', r)
 
     def test_POST_form(self):
         r = http(
@@ -431,7 +461,9 @@ class AutoContentTypeAndAcceptHeadersTest(BaseTestCase):
         )
         self.assertIn(OK, r)
         self.assertEqual(r.json['headers']['Accept'], 'application/json')
-        self.assertFalse(r.json['headers'].get('Content-Type'))
+        # Make sure Content-Type gets set even with no data.
+        # https://github.com/jkbr/httpie/issues/137
+        self.assertIn('application/json', r.json['headers']['Content-Type'])
 
     def test_GET_explicit_JSON_explicit_headers(self):
         r = http(
@@ -514,7 +546,7 @@ class ImplicitHTTPMethodTest(BaseTestCase):
             'hello=world'
         )
         self.assertIn(OK, r)
-        self.assertIn('"hello": "world"', r)
+        self.assertIn(r'\"hello\": \"world\"', r)
 
     def test_implicit_POST_form(self):
         r = http(
@@ -653,8 +685,8 @@ class VerboseFlagTest(BaseTestCase):
             'baz=bar'
         )
         self.assertIn(OK, r)
-        #noinspection PyUnresolvedReferences
-        self.assertEqual(r.count('"baz": "bar"'), 2)
+        self.assertIn('"baz": "bar"', r)  # request
+        self.assertIn(r'\"baz\": \"bar\"', r)  # response
 
 
 class MultipartFormDataFileUploadTest(BaseTestCase):
@@ -778,6 +810,7 @@ class RequestBodyFromFilePathTest(BaseTestCase):
     """
     def test_request_body_from_file_by_path(self):
         r = http(
+            '--verbose',
             'POST',
             httpbin('/post'),
             '@' + FILE_PATH_ARG
@@ -840,8 +873,8 @@ class AuthTest(BaseTestCase):
             httpbin('/digest-auth/auth/user/password')
         )
         self.assertIn(OK, r)
-        self.assertIn('"authenticated": true', r)
-        self.assertIn('"user": "user"', r)
+        self.assertIn(r'"authenticated": true', r)
+        self.assertIn(r'"user": "user"', r)
 
     def test_password_prompt(self):
 
@@ -858,6 +891,31 @@ class AuthTest(BaseTestCase):
         self.assertIn('"authenticated": true', r)
         self.assertIn('"user": "user"', r)
 
+    def test_credentials_in_url(self):
+        url = httpbin('/basic-auth/user/password')
+        url = 'http://user:password@' + url.split('http://', 1)[1]
+        r = http(
+            'GET',
+            url
+        )
+        self.assertIn(OK, r)
+        self.assertIn('"authenticated": true', r)
+        self.assertIn('"user": "user"', r)
+
+    def test_credentials_in_url_auth_flag_has_priority(self):
+        """When credentials are passed in URL and via -a at the same time,
+         then the ones from -a are used."""
+        url = httpbin('/basic-auth/user/password')
+        url = 'http://user:wrong_password@' + url.split('http://', 1)[1]
+        r = http(
+            '--auth=user:password',
+            'GET',
+            url
+        )
+        self.assertIn(OK, r)
+        self.assertIn('"authenticated": true', r)
+        self.assertIn('"user": "user"', r)
+
 
 class ExitStatusTest(BaseTestCase):
 
@@ -867,7 +925,7 @@ class ExitStatusTest(BaseTestCase):
             httpbin('/status/200')
         )
         self.assertIn(OK, r)
-        self.assertEqual(r.exit_status, exit.OK)
+        self.assertEqual(r.exit_status, ExitStatus.OK)
 
     def test_error_response_exits_0_without_check_status(self):
         r = http(
@@ -875,7 +933,7 @@ class ExitStatusTest(BaseTestCase):
             httpbin('/status/500')
         )
         self.assertIn('HTTP/1.1 500', r)
-        self.assertEqual(r.exit_status, exit.OK)
+        self.assertEqual(r.exit_status, ExitStatus.OK)
         self.assertTrue(not r.stderr)
 
     def test_timeout_exit_status(self):
@@ -884,7 +942,7 @@ class ExitStatusTest(BaseTestCase):
             'GET',
             httpbin('/delay/1')
         )
-        self.assertEqual(r.exit_status, exit.ERROR_TIMEOUT)
+        self.assertEqual(r.exit_status, ExitStatus.ERROR_TIMEOUT)
 
     def test_3xx_check_status_exits_3_and_stderr_when_stdout_redirected(self):
         r = http(
@@ -895,7 +953,7 @@ class ExitStatusTest(BaseTestCase):
             env=TestEnvironment(stdout_isatty=False,)
         )
         self.assertIn('HTTP/1.1 301', r)
-        self.assertEqual(r.exit_status, exit.ERROR_HTTP_3XX)
+        self.assertEqual(r.exit_status, ExitStatus.ERROR_HTTP_3XX)
         self.assertIn('301 moved permanently', r.stderr.lower())
 
     @skipIf(requests_version == '0.13.6',
@@ -909,7 +967,7 @@ class ExitStatusTest(BaseTestCase):
         )
         # The redirect will be followed so 200 is expected.
         self.assertIn('HTTP/1.1 200 OK', r)
-        self.assertEqual(r.exit_status, exit.OK)
+        self.assertEqual(r.exit_status, ExitStatus.OK)
 
     def test_4xx_check_status_exits_4(self):
         r = http(
@@ -918,7 +976,7 @@ class ExitStatusTest(BaseTestCase):
             httpbin('/status/401')
         )
         self.assertIn('HTTP/1.1 401', r)
-        self.assertEqual(r.exit_status, exit.ERROR_HTTP_4XX)
+        self.assertEqual(r.exit_status, ExitStatus.ERROR_HTTP_4XX)
         # Also stderr should be empty since stdout isn't redirected.
         self.assertTrue(not r.stderr)
 
@@ -929,7 +987,7 @@ class ExitStatusTest(BaseTestCase):
             httpbin('/status/500')
         )
         self.assertIn('HTTP/1.1 500', r)
-        self.assertEqual(r.exit_status, exit.ERROR_HTTP_5XX)
+        self.assertEqual(r.exit_status, ExitStatus.ERROR_HTTP_5XX)
 
 
 class WindowsOnlyTests(BaseTestCase):
@@ -982,7 +1040,8 @@ class StreamTest(BaseTestCase):
         #self.assertIn(OK_COLOR, r)
 
     def test_encoded_stream(self):
-        """Test that --stream works with non-prettified redirected terminal output."""
+        """Test that --stream works with non-prettified
+        redirected terminal output."""
         with open(BIN_FILE_PATH, 'rb') as f:
             r = http(
                 '--pretty=none',
@@ -1000,7 +1059,8 @@ class StreamTest(BaseTestCase):
         #self.assertIn(OK, r)
 
     def test_redirected_stream(self):
-        """Test that --stream works with non-prettified redirected terminal output."""
+        """Test that --stream works with non-prettified
+        redirected terminal output."""
         with open(BIN_FILE_PATH, 'rb') as f:
             r = http(
                 '--pretty=none',
@@ -1019,12 +1079,35 @@ class StreamTest(BaseTestCase):
         self.assertIn(BIN_FILE_CONTENT, r)
 
 
+class IgnoreStdinTest(BaseTestCase):
+
+    def test_ignore_stdin(self):
+        with open(FILE_PATH) as f:
+            r = http(
+                '--ignore-stdin',
+                '--verbose',
+                httpbin('/get'),
+                env=TestEnvironment(stdin=f, stdin_isatty=False)
+            )
+        self.assertIn(OK, r)
+        self.assertIn('GET /get HTTP', r)  # Don't default to POST.
+        self.assertNotIn(FILE_CONTENT, r)  # Don't send stdin data.
+
+    def test_ignore_stdin_cannot_prompt_password(self):
+        r = http(
+            '--ignore-stdin',
+            '--auth=username-without-password',
+            httpbin('/get'),
+        )
+        self.assertEqual(r.exit_status, ExitStatus.ERROR)
+        self.assertIn('because --ignore-stdin', r.stderr)
+
+
 class LineEndingsTest(BaseTestCase):
     """Test that CRLF is properly used in headers and
     as the headers/body separator."""
 
     def _validate_crlf(self, msg):
-        #noinspection PyUnresolvedReferences
         lines = iter(msg.splitlines(True))
         for header in lines:
             if header == CRLF:
@@ -1059,7 +1142,7 @@ class LineEndingsTest(BaseTestCase):
             'GET',
             httpbin('/get')
         )
-        self.assertEqual(r.exit_status,0)
+        self.assertEqual(r.exit_status, 0)
         self._validate_crlf(r)
 
     def test_CRLF_ugly_request(self):
@@ -1112,6 +1195,8 @@ class ItemParsingTest(BaseTestCase):
             # files
             self.key_value_type('bar\\@baz@%s' % FILE_PATH_ARG)
         ])
+        # `requests.structures.CaseInsensitiveDict` => `dict`
+        headers = dict(headers._store.values())
         self.assertDictEqual(headers, {
             'foo:bar': 'baz',
             'jack@jill': 'hill',
@@ -1141,6 +1226,8 @@ class ItemParsingTest(BaseTestCase):
             self.key_value_type('test-file@%s' % FILE_PATH_ARG),
             self.key_value_type('query==value'),
         ])
+        # `requests.structures.CaseInsensitiveDict` => `dict`
+        headers = dict(headers._store.values())
         self.assertDictEqual(headers, {
             'header': 'value',
             'eh': ''
@@ -1164,78 +1251,113 @@ class ArgumentParserTestCase(unittest.TestCase):
         self.parser = input.Parser()
 
     def test_guess_when_method_set_and_valid(self):
-        args = argparse.Namespace()
-        args.method = 'GET'
-        args.url = 'http://example.com/'
-        args.items = []
+        self.parser.args = argparse.Namespace()
+        self.parser.args.method = 'GET'
+        self.parser.args.url = 'http://example.com/'
+        self.parser.args.items = []
+        self.parser.args.ignore_stdin = False
 
-        self.parser._guess_method(args, TestEnvironment())
+        self.parser.env = TestEnvironment()
 
-        self.assertEqual(args.method, 'GET')
-        self.assertEqual(args.url, 'http://example.com/')
-        self.assertEqual(args.items, [])
+        self.parser._guess_method()
+
+        self.assertEqual(self.parser.args.method, 'GET')
+        self.assertEqual(self.parser.args.url, 'http://example.com/')
+        self.assertEqual(self.parser.args.items, [])
 
     def test_guess_when_method_not_set(self):
-        args = argparse.Namespace()
-        args.method = None
-        args.url = 'http://example.com/'
-        args.items = []
 
-        self.parser._guess_method(args, TestEnvironment())
+        self.parser.args = argparse.Namespace()
+        self.parser.args.method = None
+        self.parser.args.url = 'http://example.com/'
+        self.parser.args.items = []
+        self.parser.args.ignore_stdin = False
+        self.parser.env = TestEnvironment()
 
-        self.assertEqual(args.method, 'GET')
-        self.assertEqual(args.url, 'http://example.com/')
-        self.assertEqual(args.items, [])
+        self.parser._guess_method()
+
+        self.assertEqual(self.parser.args.method, 'GET')
+        self.assertEqual(self.parser.args.url, 'http://example.com/')
+        self.assertEqual(self.parser.args.items, [])
 
     def test_guess_when_method_set_but_invalid_and_data_field(self):
-        args = argparse.Namespace()
-        args.method = 'http://example.com/'
-        args.url = 'data=field'
-        args.items = []
+        self.parser.args = argparse.Namespace()
+        self.parser.args.method = 'http://example.com/'
+        self.parser.args.url = 'data=field'
+        self.parser.args.items = []
+        self.parser.args.ignore_stdin = False
+        self.parser.env = TestEnvironment()
+        self.parser._guess_method()
 
-        self.parser._guess_method(args, TestEnvironment())
-
-        self.assertEqual(args.method, 'POST')
-        self.assertEqual(args.url, 'http://example.com/')
+        self.assertEqual(self.parser.args.method, 'POST')
+        self.assertEqual(self.parser.args.url, 'http://example.com/')
         self.assertEqual(
-            args.items,
+            self.parser.args.items,
             [input.KeyValue(
                 key='data', value='field', sep='=', orig='data=field')])
 
     def test_guess_when_method_set_but_invalid_and_header_field(self):
-        args = argparse.Namespace()
-        args.method = 'http://example.com/'
-        args.url = 'test:header'
-        args.items = []
+        self.parser.args = argparse.Namespace()
+        self.parser.args.method = 'http://example.com/'
+        self.parser.args.url = 'test:header'
+        self.parser.args.items = []
+        self.parser.args.ignore_stdin = False
 
-        self.parser._guess_method(args, TestEnvironment())
+        self.parser.env = TestEnvironment()
 
-        self.assertEqual(args.method, 'GET')
-        self.assertEqual(args.url, 'http://example.com/')
+        self.parser._guess_method()
+
+        self.assertEqual(self.parser.args.method, 'GET')
+        self.assertEqual(self.parser.args.url, 'http://example.com/')
         self.assertEqual(
-            args.items,
+            self.parser.args.items,
             [input.KeyValue(
                 key='test', value='header', sep=':', orig='test:header')])
 
     def test_guess_when_method_set_but_invalid_and_item_exists(self):
-        args = argparse.Namespace()
-        args.method = 'http://example.com/'
-        args.url = 'new_item=a'
-        args.items = [
+        self.parser.args = argparse.Namespace()
+        self.parser.args.method = 'http://example.com/'
+        self.parser.args.url = 'new_item=a'
+        self.parser.args.items = [
             input.KeyValue(
                 key='old_item', value='b', sep='=', orig='old_item=b')
         ]
+        self.parser.args.ignore_stdin = False
 
-        self.parser._guess_method(args, TestEnvironment())
+        self.parser.env = TestEnvironment()
 
-        self.assertEqual(args.items, [
+        self.parser._guess_method()
+
+        self.assertEqual(self.parser.args.items, [
             input.KeyValue(
                 key='new_item', value='a', sep='=', orig='new_item=a'),
-            input.KeyValue(key
-                ='old_item', value='b', sep='=', orig='old_item=b'),
+            input.KeyValue(
+                key='old_item', value='b', sep='=', orig='old_item=b'),
         ])
 
 
+class TestNoOptions(BaseTestCase):
+
+    def test_valid_no_options(self):
+        r = http(
+            '--verbose',
+            '--no-verbose',
+            'GET',
+            httpbin('/get')
+        )
+        self.assertNotIn('GET /get HTTP/1.1', r)
+
+    def test_invalid_no_options(self):
+        r = http(
+            '--no-war',
+            'GET',
+            httpbin('/get')
+        )
+        self.assertEqual(r.exit_status, 1)
+        self.assertIn('unrecognized arguments: --no-war', r.stderr)
+        self.assertNotIn('GET /get HTTP/1.1', r)
+
+
 class READMETest(BaseTestCase):
 
     @skipIf(not has_docutils(), 'docutils not installed')
@@ -1282,6 +1404,26 @@ class SessionTest(BaseTestCase):
         self.assertEqual(r.json['headers']['Cookie'], 'hello=world')
         self.assertIn('Basic ', r.json['headers']['Authorization'])
 
+    def test_session_ignored_header_prefixes(self):
+        r = http(
+            '--session=test',
+            'GET',
+            httpbin('/get'),
+            'Content-Type: text/plain',
+            'If-Unmodified-Since: Sat, 29 Oct 1994 19:43:31 GMT',
+            env=self.env
+        )
+        self.assertIn(OK, r)
+
+        r2 = http(
+            '--session=test',
+            'GET',
+            httpbin('/get')
+        )
+        self.assertIn(OK, r2)
+        self.assertNotIn('Content-Type', r2.json['headers'])
+        self.assertNotIn('If-Unmodified-Since', r2.json['headers'])
+
     def test_session_update(self):
         # Get a response to a request from the original session.
         r1 = http(
@@ -1350,10 +1492,187 @@ class SessionTest(BaseTestCase):
         )
         self.assertIn(OK, r3)
 
+        # Origin can differ on Travis.
+        del r1.json['origin'], r3.json['origin']
+
         # Should be the same as before r2.
         self.assertDictEqual(r1.json, r3.json)
 
+    def test_session_by_path(self):
+        session_path = os.path.join(self.config_dir, 'session-by-path.json')
+
+        r1 = http(
+            '--session=' + session_path,
+            'GET',
+            httpbin('/get'),
+            'Foo:Bar',
+            env=self.env
+        )
+        self.assertIn(OK, r1)
+
+        r2 = http(
+            '--session=' + session_path,
+            'GET',
+            httpbin('/get'),
+            env=self.env
+        )
+        self.assertIn(OK, r2)
+
+        self.assertEqual(r2.json['headers']['Foo'], 'Bar')
+
+
+class DownloadUtilsTest(BaseTestCase):
+
+    def test_Content_Range_parsing(self):
+
+        parse = parse_content_range
+
+        self.assertEqual(parse('bytes 100-199/200', 100), 200)
+        self.assertEqual(parse('bytes 100-199/*', 100), 200)
+
+        # missing
+        self.assertRaises(ContentRangeError, parse, None, 100)
+
+        # syntax error
+        self.assertRaises(ContentRangeError, parse, 'beers 100-199/*', 100)
+
+        # unexpected range
+        self.assertRaises(ContentRangeError, parse, 'bytes 100-199/*', 99)
+
+        # invalid instance-length
+        self.assertRaises(ContentRangeError, parse, 'bytes 100-199/199', 100)
+
+        # invalid byte-range-resp-spec
+        self.assertRaises(ContentRangeError, parse, 'bytes 100-99/199', 100)
+
+        # invalid byte-range-resp-spec
+        self.assertRaises(ContentRangeError, parse, 'bytes 100-100/*', 100)
+
+    def test_Content_Disposition_parsing(self):
+        parse = filename_from_content_disposition
+        self.assertEqual(
+            parse('attachment; filename=hello-WORLD_123.txt'),
+            'hello-WORLD_123.txt'
+        )
+        self.assertEqual(
+            parse('attachment; filename=".hello-WORLD_123.txt"'),
+            'hello-WORLD_123.txt'
+        )
+        self.assertEqual(
+            parse('attachment; filename="white space.txt"'),
+            'white space.txt'
+        )
+        self.assertEqual(
+            parse(r'attachment; filename="\"quotes\".txt"'),
+            '"quotes".txt'
+        )
+        self.assertEqual(parse('attachment; filename=/etc/hosts'), 'hosts')
+        self.assertIsNone(parse('attachment; filename='))
+
+    def test_filename_from_url(self):
+        self.assertEqual(filename_from_url(
+            url='http://example.org/foo',
+            content_type='text/plain'
+        ), 'foo.txt')
+        self.assertEqual(filename_from_url(
+            url='http://example.org/foo',
+            content_type='text/html; charset=utf8'
+        ), 'foo.html')
+        self.assertEqual(filename_from_url(
+            url='http://example.org/foo',
+            content_type=None
+        ), 'foo')
+        self.assertEqual(filename_from_url(
+            url='http://example.org/foo',
+            content_type='x-foo/bar'
+        ), 'foo')
+
+    def test_unique_filename(self):
+
+        def make_exists(unique_on_attempt=0):
+            # noinspection PyUnresolvedReferences,PyUnusedLocal
+            def exists(filename):
+                if exists.attempt == unique_on_attempt:
+                    return False
+                exists.attempt += 1
+                return True
+            exists.attempt = 0
+            return exists
+
+        self.assertEqual(
+            get_unique_filename('foo.bar', exists=make_exists()),
+            'foo.bar'
+        )
+        self.assertEqual(
+            get_unique_filename('foo.bar', exists=make_exists(1)),
+            'foo.bar-1'
+        )
+        self.assertEqual(
+            get_unique_filename('foo.bar', exists=make_exists(10)),
+            'foo.bar-10'
+        )
+
+
+class Response(object):
+
+    # noinspection PyDefaultArgument
+    def __init__(self, url, headers={}, status_code=200):
+        self.url = url
+        self.headers = CaseInsensitiveDict(headers)
+        self.status_code = status_code
+
+
+# noinspection PyTypeChecker
+class DownloadTest(BaseTestCase):
+    # TODO: more tests
+
+    def test_actual_download(self):
+        url = httpbin('/robots.txt')
+        body = urlopen(url).read().decode()
+        r = http(
+            '--download',
+            url,
+            env=TestEnvironment(
+                stdin_isatty=True,
+                stdout_isatty=False
+            )
+        )
+        self.assertIn('Downloading', r.stderr)
+        self.assertIn('[K', r.stderr)
+        self.assertIn('Done', r.stderr)
+        self.assertEqual(body, r)
+
+    def test_download_with_Content_Length(self):
+        download = Download(output_file=open(os.devnull, 'w'))
+        download.start(Response(
+            url=httpbin('/'),
+             headers={'Content-Length': 10}
+        ))
+        time.sleep(1.1)
+        download.chunk_downloaded(b'12345')
+        time.sleep(1.1)
+        download.chunk_downloaded(b'12345')
+        download.finish()
+        self.assertFalse(download.interrupted)
+
+    def test_download_no_Content_Length(self):
+        download = Download(output_file=open(os.devnull, 'w'))
+        download.start(Response(url=httpbin('/')))
+        time.sleep(1.1)
+        download.chunk_downloaded(b'12345')
+        download.finish()
+        self.assertFalse(download.interrupted)
+
+    def test_download_interrupted(self):
+        download = Download(output_file=open(os.devnull, 'w'))
+        download.start(Response(
+            url=httpbin('/'),
+            headers={'Content-Length': 5}
+        ))
+        download.chunk_downloaded(b'1234')
+        download.finish()
+        self.assertTrue(download.interrupted)
+
 
 if __name__ == '__main__':
-    #noinspection PyCallingNonCallable
     unittest.main()
